<!DOCTYPE html>
<html>
    <body>
        <h1>Form</h1>
        {% with messages = get_flashed_messages() %}
            {% if messages %}
            <ul class="messages">
                {% for message in messages %}
                <li>{{ message }}</li>
                {% endfor %}
            </ul>
            {% endif %}
        {% endwith %}
        <form id="contact-form" action="" method="post">
            <fieldset>
                <p>
                    <label for="text">text</label>
                    <input type="text" id="text" name="text" />
                </p>
                <p>
                    <label for="email">email</label>
                    <input type="email" id="email" name="email" />
                </p>
                <p>
                    <label for="textarea">textarea</label>
                    <textarea id="textarea" name="textarea"></textarea>
                </p>
                <p>
                    <label for="checkbox">checkbox</label>
                    <input type="checkbox" id="checkbox" name="checkbox" />
                </p>
                <p>
                    <label for="multiple-checkbox-first">multiple checkbox first choice</label>
                    <input type="checkbox" id="multiple-checkbox-first" name="multiple-checkbox" value="first choice" />
                    <label for="multiple-checkbox-second">multiple checkbox second choice</label>
                    <input type="checkbox" id="multiple-checkbox-second" name="multiple-checkbox" value="second choice" />
                </p>
                <p>
                    <label for="selectbox">selectbox</label>
                    <select id="selectbox" name="selectbox">
                        <option value="one">one</option>
                        <option value="two" selected="selected">Two</option>
                    </select>
                </p>
                <p>
                    <label for="selectbox">long selectbox</label>
                    <select id="long_selectbox" name="long_selectbox">
                        <option value='0'>0</option>
                        <option value='1'>1</option>
                        <option value='2'>2</option>
                        <option value='3'>3</option>
                        <option value='4'>4</option>
                        <option value='5'>5</option>
                        <option value='6'>6</option>
                        <option value='7'>7</option>
                        <option value='8'>8</option>
                        <option value='9'>9</option>
                        <option value='10'>10</option>
                        <option value='11'>11</option>
                        <option value='12'>12</option>
                        <option value='13'>13</option>
                        <option value='14'>14</option>
                        <option value='15'>15</option>
                        <option value='16'>16</option>
                        <option value='17'>17</option>
                        <option value='18'>18</option>
                        <option value='19'>19</option>
                        <option value='20'>20</option>
                        <option value='21'>21</option>
                        <option value='22'>22</option>
                        <option value='23'>23</option>
                        <option value='24'>24</option>
                        <option value='25'>25</option>
                        <option value='26'>26</option>
                        <option value='27'>27</option>
                        <option value='28'>28</option>
                        <option value='29'>29</option>
                        <option value='30'>30</option>
                        <option value='31'>31</option>
                        <option value='32'>32</option>
                        <option value='33'>33</option>
                        <option value='34'>34</option>
                        <option value='35'>35</option>
                        <option value='36'>36</option>
                        <option value='37'>37</option>
                        <option value='38'>38</option>
                        <option value='39'>39</option>
                        <option value='40'>40</option>
                        <option value='41'>41</option>
                        <option value='42'>42</option>
                        <option value='43'>43</option>
                        <option value='44'>44</option>
                        <option value='45'>45</option>
                        <option value='46'>46</option>
                        <option value='47'>47</option>
                        <option value='48'>48</option>
                        <option value='49'>49</option>
                    </select>
                </p>
                <p>
                    <label for="selectbox">multiple selectbox</label>
                    <select id="multiple_selectbox" name="multiple_selectbox" multiple="multiple" size="2">
                        <option value="one">one</option>
                        <option value="two">two</option>
                        <option value="three" selected="selected">three</option>
                        <option value="four">four</option>
                    </select>
                </p>
                <p>
                    <label for="radio-first">first choice</label>
                    <input id="radio-first" type="radio" name="radio" value="first choice" />
                    <label for="radio-second">second choice</label>
                    <input id="radio-second" type="radio" name="radio" value="second choice" />
                </p>
                <input type="submit" value="go!" />
            </fieldset>
        </form>
    </body>
</html>
