<!DOCTYPE html>
<html>
    <body>
        <h1>Form</h1>
        <form id="contact-form" action="/upload" method="post" enctype="multipart/form-data">
            <fieldset>
                <p>
                    <label for="simple-file">simple file upload</label>
                    <input id="simple-file" type="file" name="simple-file" />
                </p>
                <input type="submit" value="go!" />
            </fieldset>
        </form>
    </body>
</html>