<!DOCTYPE html>
<html>
    <body>
        <h1>Test page</h1>
        <p>This is the home page.</p>
        <p><a href="{{ url_for('form') }}">This is a clickable link</a></p>
        <script type="text/javascript">
            var myGlobal = 'my global';
        </script>
        <div id="top">
            I'm top.
        </div>

        <div id="middle" style="position: absolute;left: 1000px;top: 1000px">
            I'm middle.
        </div>

        <div id="right" style="position: absolute;left: 2000px;top: 1000px">
            I'm right.
        </div>

        <div id="bottom" style="position: absolute;left: 1000px;top: 2000px">
            I'm bottom.
        </div>
    </body>
</html>