<!DOCTYPE html>
<html>
    <body>
        <h1>Test page</h1>
        <p>This is the home page.</p>
        <p><a href="{{ url_for('form') }}">This is a clickable link</a></p>
        <script type="text/javascript">
            document.location.hash = "test";
        </script>
    </body>
</html>